from ..mesh_data import RawMeshData
from ..mesh_attributes import Attribute
from ...geometry import Vec
import numpy as np
from enum import Enum
from ...config import NOT_AN_ID

class Chunk:

    class Type(Enum):
        HEAD = 0
        ATTR = 1
        ATTS = 2

        @classmethod
        def from_string(cls, txt : str):
            if txt=="[HEAD]": return cls.HEAD
            if txt=="[ATTR]": return cls.ATTR
            if txt=="[ATTS]": return cls.ATTS

    class Container(Enum):
        VERTICES = 0
        EDGES = 1
        FACES = 2
        FACE_CORNERS = 3
        CELLS = 4
        CELL_CORNERS = 5
        CELL_FACETS = 6

        @classmethod
        def from_string(cls, txt : str):
            if "vertices" in txt: return cls.VERTICES
            if "edges" in txt : return cls.EDGES
            if "facet_corners" in txt : return cls.FACE_CORNERS
            if "facets" in txt : return cls.FACES
            if "cell_corners" in txt : return cls.CELL_CORNERS
            if "cell_facets" in txt : return cls.CELL_FACETS
            if "cells" in txt : return cls.CELLS

        def to_string(self):
            return { 
                0 : "\"GEO::Mesh::vertices\"",
                1 : "\"GEO::Mesh::edges\"",
                2 : "\"GEO::Mesh::facets\"",
                3 : "\"GEO::Mesh::facet_corners\"",
                4 : "\"GEO::Mesh::cells\"",
                5 : "\"GEO::Mesh::cell_corners\"",
                6 : "\"GEO::Mesh::cell_facets\""
            }.get(self.value, None)

    def __init__(self, chunk_data):
        self.type : Chunk.Type = Chunk.Type.from_string(chunk_data[0])
        self.container : Chunk.Container = Chunk.Container.from_string(chunk_data[1]) # like GEO::Mesh::vertices
        
        if self.type == Chunk.Type.ATTR:
            self.name : str = chunk_data[2]
            self.data_type : Attribute.Type = Attribute.Type.from_string(chunk_data[3])
            self.data_size : int = int(chunk_data[4]) # number of data point per element
            self.n_data : int = int(chunk_data[5]) # number of elements
            if self.data_type == Attribute.Type.Float:
                self.data = [np.float64(x) for x in chunk_data[6:]]
            elif self.data_type == Attribute.Type.Int:
                self.data = [int(x) for x in chunk_data[6:]]
            elif self.data_type == Attribute.Type.Bool:
                self.data = [bool(int(x)) for x in chunk_data[6:]]
            else:
                # Attribute type cannot be exported
                self.data = chunk_data[6:]

        elif self.type == Chunk.Type.ATTS:
            self.n : int = int(chunk_data[2])

def is_chunk_header(line : str) -> bool:
    return "[HEAD]" in line or "[ATTS]" in line or "[ATTR]" in line
    # ignore other type of blocks (parameters for geogram / graphite)

def import_attribute(chk : Chunk, attr: Attribute):
    for i in range(len(chk.data)//chk.n_data):
        val = []
        for j in range(chk.n_data):
            val.append(chk.data[chk.n_data*i + j])
        if chk.n_data==1 and val[0]!= attr.default_value:
            attr[i] = val[0]
        elif chk.n_data>1:
            attr[i] = val

def import_geogram_ascii(path):

    data = None
    with open(path, 'r') as f:
        data = [x.split("#")[0].strip() for x in f.readlines()] # split to remove comments

    # Detect chunk separators
    chunk_sep = [] # int values for header lines -> separators of data
    for i,line in enumerate(data):
        if is_chunk_header(line):
            chunk_sep.append(i)
    chunk_sep.append(len(data))

    # build chunk objects
    chunks = []
    for i in range(len(chunk_sep)-1):
        start =  chunk_sep[i]
        end = chunk_sep[i+1]
        chunk_data = []
        for k in range(start,end):
            chunk_data.append(data[k])
        chunks.append(chunk_data)
    chunks = [Chunk(data) for data in chunks]

    outmesh = RawMeshData()

    # First read ATTS chunks
    container_sizes = dict([(a,0) for a in Chunk.Container])
    for chk in chunks:
        if chk.type != Chunk.Type.ATTS: continue
        container_sizes[chk.container] = chk.n

    # Build facet index
    n_corner_in_facet = []
    facet_ptr = []

    n_corner_in_cell = []
    cell_ptr = []
    for chk in chunks:
        if chk.type == Chunk.Type.ATTR and chk.name == "\"GEO::Mesh::facets::facet_ptr\"":
            # facet sizes are provided : the mesh is not triangular
            for i in range(container_sizes[Chunk.Container.FACES]-1):
                n_corner_in_facet.append(chk.data[i+1] - chk.data[i])
            n_corner_in_facet.append(container_sizes[Chunk.Container.FACE_CORNERS] - chk.data[-1])
            facet_ptr = chk.data

        elif chk.type == Chunk.Type.ATTR and chk.name == "\"GEO::Mesh::cells::cell_ptr\"":
            # cell sizrs are provided : the mesh is not tetrahedral
            for i in range(container_sizes[Chunk.Container.CELLS]-1):
                n_corner_in_cell.append(chk.data[i+1] - chk.data[i])
            n_corner_in_facet.append(container_sizes[Chunk.Container.CELL_CORNERS] - chk.data[-1])
            cell_ptr = chk.data

    if len(n_corner_in_facet)==0 and container_sizes[Chunk.Container.FACES]>0:
        # We have faces but no facet_ptr chunk. By convention, all faces are triangles
        n_corner_in_facet = [3]*container_sizes[Chunk.Container.FACES]
        ptr = 0
        facet_ptr = []
        for c in n_corner_in_facet:
            facet_ptr.append(ptr)
            ptr += c

    if len(n_corner_in_cell)==0 and container_sizes[Chunk.Container.CELLS]>0:
        # We have cells but no cell_ptr chunk. By convention, all cells are tetrahedra
        n_corner_in_cell = [4]*container_sizes[Chunk.Container.CELLS]
        ptr = 0
        cell_ptr = []
        for c in n_corner_in_cell:
            cell_ptr.append(ptr)
            ptr += c
    
    # Build mesh data from chunk objects
    for chk in chunks:

        if chk.type == Chunk.Type.ATTS or chk.type == Chunk.Type.HEAD: continue # already treated and HEAD is ignored
        # chk.type is now [ATTR]

        # first handle special cases : vertices coordinates, edges, faces and cells connectivity
        if chk.container == Chunk.Container.VERTICES and chk.name == "\"point\"":
            assert chk.n_data==3 # points have three coordinates
            for i in range(len(chk.data)//chk.n_data):
                outmesh.vertices.append(Vec([chk.data[3*i], chk.data[3*i+1], chk.data[3*i+2] ]))
        
        elif chk.container == Chunk.Container.EDGES and chk.name == "\"GEO::Mesh::edges::edge_vertex\"":
            assert chk.n_data==2 # edges should have 2 pointers to vertices id
            for i in range(container_sizes[Chunk.Container.EDGES]):
                outmesh.edges.append([chk.data[2*i], chk.data[2*i+1]])
        
        elif chk.container == Chunk.Container.FACE_CORNERS and chk.name == "\"GEO::Mesh::facet_corners::corner_vertex\"":
            assert chk.n_data==1
            for i in range(container_sizes[Chunk.Container.FACES]):
                ncf = n_corner_in_facet[i]
                ptr = facet_ptr[i]
                face = [chk.data[ptr+_i] for _i in range(ncf)]
                outmesh.faces.append(face)
                outmesh.face_corners += [(x,i) for x in face]

        elif chk.container == Chunk.Container.FACE_CORNERS and chk.name == "\"GEO::Mesh::facet_corners::corner_adjacent_facet\"":
            assert chk.n_data==1
            adj_corner = outmesh.face_corners.create_attribute("opposite_face", int, default_value=NOT_AN_ID)
            import_attribute(chk, adj_corner)

        elif chk.container == Chunk.Container.CELL_CORNERS and chk.name == "\"GEO::Mesh::cell_corners::corner_vertex\"":
            assert chk.n_data==1
            for i in range(container_sizes[Chunk.Container.CELLS]):
                ncc = n_corner_in_cell[i]
                ptr = cell_ptr[i]
                cell = [chk.data[ptr+_i] for _i in range(ncc)]
                outmesh.cells.append(cell)
                outmesh.cell_corners += [(x,i) for x in cell]

        elif chk.container == Chunk.Container.CELL_FACETS and chk.name == "\"GEO::Mesh::cell_facets::adjacent_cell\"":
            assert chk.n_data==1
            # chk.name = "adjacent_cell"
            adj_cell = outmesh.cell_faces.create_attribute("opposite_cell", int, default_value=NOT_AN_ID)
            adj_cell._expand(container_sizes[Chunk.Container.CELL_FACETS])
            import_attribute(chk, adj_cell)

        else: # user defined attribute
            container = {
                Chunk.Container.VERTICES : outmesh.vertices,
                Chunk.Container.EDGES : outmesh.edges,
                Chunk.Container.FACES : outmesh.faces,
                Chunk.Container.FACE_CORNERS : outmesh.face_corners,
                Chunk.Container.CELLS : outmesh.cells,
                Chunk.Container.CELL_CORNERS : outmesh.cell_corners,
                Chunk.Container.CELL_FACETS : outmesh.cell_faces,
            }.get(chk.container, None)
            if container is None:
                err_msg = f"In import_geogram_ascii : Container {chk.container} is not recognized"
                raise Exception(err_msg)
            attr = container.create_attribute(chk.name.split("\"")[1], chk.data_type, chk.n_data) # remove first and last "
            if chk.container == Chunk.Container.CELL_FACETS:
                attr._expand(container_sizes[Chunk.Container.CELL_FACETS])
            import_attribute(chk, attr)
    return outmesh

def export_attribute(f, size, container, attr, attr_name):
    f.write(f"[ATTR]\n\"{container}\"\n\"{attr_name}\"\n\"{attr.type.to_string()}\"\n{attr.type.byte_size()}\n{attr.elemsize}\n")
    for i in range(size):
        if attr.elemsize==1:
            if attr.type==Attribute.Type.Bool: # should be written as 0 or 1 and not as "true" or "false"
                f.write(f"{int(attr[i])}\n")
            else:
                f.write("{}\n".format(attr[i]))
        else:
            for j in range(attr.elemsize):
                if attr.type==Attribute.Type.Bool:
                    f.write(f"{int(attr[i][j])}\n")
                else:
                    f.write(f"{attr[i][j]}\n")

def export_geogram_ascii(mesh : RawMeshData, path):
    with open(path, "w", newline="\n") as f:
        f.write("[HEAD]\n\"GEOGRAM\"\n\"1.0\"\n")
        
        # Vertices
        n_vert = len(mesh.vertices)
        f.write("[ATTS]\n\"GEO::Mesh::vertices\"\n{}\n".format(n_vert))
        f.write("[ATTR]\n\"GEO::Mesh::vertices\"\n\"point\"\n\"double\"\n8\n3\n")
        for i in range(n_vert):
            f.write("{}\n{}\n{}\n".format(*mesh.vertices[i]))
        for attr_key in mesh.vertices.attributes:
            attr = mesh.vertices.get_attribute(attr_key)
            export_attribute(f, n_vert, "GEO::Mesh::vertices", attr, attr_key)
 
        # Edges
        if hasattr(mesh, "edges") and not mesh.edges.empty():
            n_edges = len(mesh.edges)
            f.write("[ATTS]\n\"GEO::Mesh::edges\"\n{}\n".format(n_edges))
            f.write("[ATTR]\n\"GEO::Mesh::edges\"\n\"GEO::Mesh::edges::edge_vertex\"\n\"index_t\"\n4\n2\n")
            for edge in mesh.edges:
                f.write(f"{edge[0]}\n{edge[1]}\n")
            for attr_key in mesh.edges.attributes:
                attr = mesh.edges.get_attribute(attr_key)
                export_attribute(f, n_edges, "GEO::Mesh::edges", attr, attr_key)

        # faces
        if hasattr(mesh, "faces") and not mesh.faces.empty():
            n_face = len(mesh.faces)
            f.write(f"[ATTS]\n\"GEO::Mesh::facets\"\n{n_face}\n")
            for attr_key in mesh.faces.attributes:
                attr = mesh.faces.get_attribute(attr_key)
                export_attribute(f, n_face, "GEO::Mesh::facets", attr, attr_key)

            # face_corners
            n_corners = len(mesh.face_corners)
            f.write("[ATTS]\n\"GEO::Mesh::facet_corners\"\n{}\n".format(n_corners))
            f.write("[ATTR]\n\"GEO::Mesh::facet_corners\"\n\"GEO::Mesh::facet_corners::corner_vertex\"\n\"index_t\"\n4\n1\n")
            for c in mesh.face_corners:
                f.write(f"{c}\n")

            if mesh.face_corners.has_attribute("corner_adjacent_facet"):
                f.write("[ATTR]\n\"GEO::Mesh::facet_corners\"\n\"GEO::Mesh::facet_corners::corner_adjacent_facet\"\n\"index_t\"\n4\n1\n")
                corner_adjacent_face = mesh.face_corners.get_attribute('corner_adjacent_facet')
                for i in range(n_corners):
                    f.write(f"{corner_adjacent_face[i]}\n")
                
            for attr_key in mesh.face_corners.attributes:
                if attr_key == "corner_adjacent_facet" : continue # already handled
                attr = mesh.face_corners.get_attribute(attr_key)
                export_attribute(f, n_corners, "GEO::Mesh::facet_corners", attr, attr_key)

        # Cells
        if hasattr(mesh, "cells") and not mesh.cells.empty():
            n_cells = len(mesh.cells)
            f.write("[ATTS]\n\"GEO::Mesh::cells\"\n{}\n".format(n_cells))
            for attr_key in mesh.cells.attributes:
                attr = mesh.cells.get_attribute(attr_key)
                export_attribute(f, n_cells, "GEO::Mesh::cells", attr, attr_key)

            # Cell Corners
            n_corners = sum([len(cell) for cell in mesh.cells])
            f.write("[ATTS]\n\"GEO::Mesh::cell_corners\"\n{}\n".format(n_corners))
            f.write("[ATTR]\n\"GEO::Mesh::cell_corners\"\n\"GEO::Mesh::cell_corners::corner_vertex\"\n\"index_t\"\n4\n1\n")
            for cell in mesh.cells:
                for x in cell:
                    f.write(f"{x}\n")
            for attr_key in mesh.cell_corners.attributes:
                attr = mesh.cell_corners.get_attribute(attr_key)
                export_attribute(f, n_corners, "GEO::Mesh::cell_corners", attr, attr_key)
                   
            # Cell faces
            n_cell_faces = sum([len(c) for c in mesh.cells])
            cell_adj = mesh.cell_faces.get_attribute("adjacent_cell")
            f.write("[ATTR]\n\"GEO::Mesh::cell_corners\"\n\"GEO::Mesh::cell_faces::adjacent_cell\"\n\"index_t\"\n4\n1\n")
            for x in cell_adj:
                f.write(f"{x}\n")

            for attr_key in mesh.cell_faces.attributes:
                if attr_key=="adjacent_cell" : continue
                attr = mesh.cell_faces.get_attribute(attr_key)
                export_attribute(f, n_cell_faces, "GEO::Mesh::cell_faces", attr, attr_key)


import numpy as np
import cmath

from ..mesh.mesh_attributes import ArrayAttribute, Attribute
from ..mesh.datatypes import *
from .. import geometry as geom
from ..geometry import Vec

@allowed_mesh_types(SurfaceMesh, VolumeMesh)
def face_area(mesh : Mesh, name="area", persistent:bool=True, dense:bool=True) -> Attribute:
    """
    Computes the barycenter point of each face.

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "area".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: a float per faces
    """
    if persistent:
        area = mesh.faces.create_attribute(name, float, dense=dense)
    else:
        area = ArrayAttribute(float, len(mesh.faces)) if dense else Attribute(float)
    for T in mesh.id_faces:
        pts = [mesh.vertices[u] for u in mesh.faces[T]]
        npt = len(pts)
        if npt==3:
            area[T] = geom.triangle_area(*pts)
        elif npt==4:
            area[T] = geom.quad_area(*pts)
        else:
            bary = sum(pts)/npt
            for i in range(npt):
                A = pts[i]
                B = pts[(i+1)%npt]
                area[T] += geom.triangle_area(A,B,bary)            
    return area

@allowed_mesh_types(SurfaceMesh)
def face_normals(mesh : SurfaceMesh, name="normals", persistent:bool=True, dense:bool=True) -> Attribute:
    """
    Computes the barycenter point of each face.

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "normals".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: 3 floats per faces
    """
    if persistent :
        normals = mesh.faces.create_attribute(name, float, 3)
    else:
        normals = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float)
    for iT,T in enumerate(mesh.faces):
        pA,pB,pC = (mesh.vertices[u] for u in T[:3])
        normals[iT] = Vec.normalized(geom.cross(pB-pA, pC-pA))
    return normals

@allowed_mesh_types(SurfaceMesh, VolumeMesh)
def face_barycenter(mesh : SurfaceMesh, name="barycenter", persistent:bool = True, dense:bool = True) -> Attribute:
    """
    Computes the barycenter point of each face.

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: one 3D vector per face
    """
    if persistent :
        bary = mesh.faces.create_attribute(name, float, 3)
    else:
        bary = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float, 3)
    for iT,T in enumerate(mesh.faces):
        bary[iT] = sum(mesh.vertices[u] for u in T) / len(T)
    return bary

@allowed_mesh_types(SurfaceMesh, VolumeMesh)
def face_circumcenter(mesh : SurfaceMesh, name="circumcenter", persistent:bool=True, dense:bool=True) -> Attribute:
    """
    Computes the circumcenter point of each triangular face.

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to True

    Returns:
        Attribute: one float per face

    Raises:
        Exception: fails if a face of the mesh is not a triangle
    """
    if persistent :
        circum = mesh.faces.create_attribute(name, float, 3)
    else:
        circum = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float, 3)
    for iF,F in enumerate(mesh.faces):
        if len(F)!=3:
            raise Exception("circumcenters can be computed only for triangular faces. Received face n°{} containing {}!=3 vertices".format(iF,len(F)))
        circum[iF] = geom.circumcenter(*(mesh.vertices[u] for u in F))
    return circum

@allowed_mesh_types(SurfaceMesh)
def faces_near_border(mesh : SurfaceMesh, dist:int = 2, name = "near_border", persistent:bool = True, dense:bool = False) -> Attribute:
    """Returns the faces that are at most at 'dist' neighbours from the boundary.
    Proceeds by region growing, starting from all faces touching the boundary and going inwards.

    Parameters:
        mesh (SurfaceMesh): the input mesh
        dist (int, optional): Extend to which we flag the faces. All faces with a path of length < dist to the boundary will be returned. Defaults to 2.
        name (str, optional): Name given to the attribute. Defaults to "barycenter".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to False

    Returns:
        Attribute: one bool per faces. If not persistent, returns a set.
    """
    
    near = set()
    for e in mesh.boundary_edges:
        T1,T2 = mesh.connectivity.edge_to_faces(*mesh.edges[e])
        if T1 is None: near.add(T2)
        else: near.add(T1)
    for _ in range(dist-1):
        new_near = set()
        for T in mesh.id_faces:
            if T in near : continue
            for T2 in mesh.connectivity.face_to_faces(T):
                if T2 in near: new_near.add(T)
        near = near | new_near
    
    if persistent:
        near_attr = mesh.faces.create_attribute(name, bool, dense=dense)
    else:
        near_attr = ArrayAttribute(bool, len(mesh.faces)) if dense else Attribute(bool)
    for f in near: near_attr[f] = True
    return near_attr

@allowed_mesh_types(SurfaceMesh)
def triangle_aspect_ratio(mesh : SurfaceMesh, name : str="aspect_ratio", persistent : bool = True, dense : bool =True) -> Attribute:
    """Computes the aspect ratio of every triangular faces. Sets aspect ratio to -1 for every other faces

    Parameters:
        mesh (SurfaceMesh): the input mesh
        name (str, optional): Name given to the attribute. Defaults to "aspect_ratio".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to False.

    Returns:
        Attribute: One float per face.
    """
    if persistent :
        ratio = mesh.faces.create_attribute(name, float, 3)
    else:
        ratio = ArrayAttribute(float, len(mesh.faces), 3) if dense else Attribute(float, 3)
    for iF,F in enumerate(mesh.faces):
        if len(F)!=3: ratio[iF] = -1 
        ratio[iF] = geom.aspect_ratio(*(mesh.vertices[u] for u in F))

@allowed_mesh_types(SurfaceMesh)
def parallel_transport_curvature(mesh : SurfaceMesh, PT:"SurfaceConnectionFaces", name : str="curvature", persistent : bool=True, dense : bool = True):
    """
    Compute the curvature of each face associated to a given parallel transport pT

    Args:
        mesh (SurfaceMesh): _description_
        PT (dict): parallel transport. Dictionnary of keys (A,B) -> direction (angle) of edge (A,B) in local basis of A
        name (str, optional): Name given to the attribute.. Defaults to "curvature".
        persistent (bool, optional): If the attribute is persistent (stored in the mesh object) or not. Defaults to True.
        dense (bool, optional): Is the attribute dense (numpy array) or not (dict). Defaults to False.
    
    Returns:
        Attribute: One float per face.
    """
    if persistent :
        curv = mesh.faces.create_attribute(name, float)
    else:
        curv = ArrayAttribute(float, len(mesh.faces)) if dense else Attribute(float)
    for iF, (A,B,C) in enumerate(mesh.faces):
        v = 1+0j
        for a,b in [(A,B) ,(B,C), (C,A)]:
            v *= cmath.rect(1., PT.transport(b,a) - PT.transport(a,b) - np.pi)
        curv[iF] = cmath.phase(v)
    return curv


from random import randint
from collections import deque

from ...mesh.datatypes import *
from ...mesh import merge
from ...attributes.misc_faces import face_barycenter
from ...utils import keyify
from .base import SpanningTree, SpanningForest

class FaceSpanningTree(SpanningTree):
    """
    A spanning tree defined over the connectivity of a mesh, built over dual edges.

    Warning:
        This tree considers a unique starting point and stops when all reachable faces are visited. 
        If the mesh is disconnected, the tree will be incomplete. In this case, use `FaceSpanningForest` class instead.
    """

    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, starting_face=None, forbidden_edges:set = None):
        super().__init__(mesh)
        if starting_face is not None:
            self.root = starting_face
        else:
            self.root = randint(0,len(self.mesh.faces)-1)
        
        if forbidden_edges is None:
            self.forbidden_edges : set = set()
        else:
            self.forbidden_edges : set = forbidden_edges
        self.parent = [None]*len(self.mesh.faces)
        self.children = [[] for v in self.mesh.id_faces]
        self.edges = []

    def compute(self):
        dist_to_root = [float("inf") for v in self.mesh.id_faces]
        seen = [False for v in self.mesh.id_faces]
        queue = deque()

        def put_neighbours_in_queue(f):
            for e in self.mesh.connectivity.face_to_edges(f):
                if e not in self.forbidden_edges:
                    a,b = self.mesh.edges[e]
                    nf = self.mesh.connectivity.opposite_face(a,b,f)
                    if (nf is not None) and (not seen[nf]):
                        queue.append((f,nf))
        
        put_neighbours_in_queue(self.root)
        dist_to_root[self.root] = 0
        seen[self.root] = True

        while len(queue):
            f,nf = queue.popleft()
            if seen[nf] : continue
            seen[nf] = True
            if dist_to_root[f] + 1 < dist_to_root[nf]:
                self.parent[nf] = f
                dist_to_root[nf] = dist_to_root[f] +1
            put_neighbours_in_queue(nf)

        # build children data from parent data
        for f in self.mesh.id_faces:
            p = self.parent[f]
            if p is not None:
                self.children[p].append(f)
                self.edges.append( keyify(p,f))
        super().compute() # sets the 'computed' flag

    def build_tree_as_polyline(self):
        """Builds the tree as a new polyline object. Useful for debug and visualization purposes

        Returns:
            PolyLine: the tree
        """
        output = PolyLine()
        if self.mesh.faces.has_attribute("barycenter"):
            bary = self.mesh.faces.get_attribute("barycenter")
        else:
            bary = face_barycenter(self.mesh)
        for iF in self.mesh.id_faces:
            output.vertices.append(bary[iF])
            if self.parent[iF] is not None:
                output.edges.append([iF, self.parent[iF]])
        return output

class FaceSpanningForest(SpanningForest):
    """A spanning forest that runs on dual edges.
    Unlike a spanning tree, will create new roots and expand trees until all faces of the mesh have been visited
    """
    
    @allowed_mesh_types(SurfaceMesh)
    def __init__(self, mesh : SurfaceMesh, forbidden_edges:set = None):
        super().__init__(mesh)
        self.forbidden_edges : set = forbidden_edges

    def compute(self) -> None :
        visited = [False]*len(self.mesh.faces)
        for f in self.mesh.id_faces:
            if not visited[f]:
                # create a tree starting from v
                self.roots.append(f)
                tree_v = FaceSpanningTree(self.mesh, f, self.forbidden_edges)()
                self.trees.append(tree_v)
                for (node, _) in tree_v.traverse():
                    visited[node] = True
        super().compute() # sets the 'computed' flag


from ..mesh_data import RawMeshData

class Mesh: 
    
    def __init__(self, dim : int, data : RawMeshData = None):
        if data is None:
            data = RawMeshData()
        else:
            data.prepare()

        self.vertices = data.vertices

        if dim>0:
            self.edges = data.edges
        
        if dim>1:
            self.faces = data.faces
            self.face_corners = data.face_corners

        if dim>2:
            self.cells = data.cells
            self.cell_corners = data.cell_corners
            self.cell_faces = data.cell_faces
